import importlib.metadata
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Label
from textual.containers import Container, Grid

class HelpModal(ModalScreen[None]):
    """Modal screen giving help/keybindings."""

    LOGO = r"""
████████╗██╗    ██╗██╗ ██████╗ 
╚══██╔══╝██║    ██║██║██╔════╝ 
   ██║   ██║ █╗ ██║██║██║  ███╗
   ██║   ██║███╗██║██║██║   ██║
   ██║   ╚███╔███╔╝██║╚██████╔╝
   ╚═╝    ╚══╝╚══╝ ╚═╝ ╚═════╝ 
"""

    CSS = """
    HelpModal {
        align: center middle;
    }

    #help-container {
        width: 70;
        height: auto;
        background: $surface;
        border: tall $primary; 
        padding: 1 2;
        align: center middle;
    }

    #help-logo {
        width: 100%;
        text-align: center;
        color: $warning;
        margin-top: 1;
        margin-bottom: 1;
    }

    .help-tagline {
        width: 100%;
        text-align: center;
        color: $success;
        text-style: bold;
        margin-bottom: 1;
    }

    .help-version {
        width: 100%;
        text-align: center;
        color: $text-muted;
        margin-bottom: 2;
    }

    #help-link {
        width: 100%;
        text-align: center;
        color: $primary;
        text-style: underline;
        margin-bottom: 1;
    }

    #coffee-link {
        width: 100%;
        text-align: center;
        color: $accent;
        margin-bottom: 2;
    }

    #help-title {
        width: 100%;
        text-align: center;
        text-style: bold underline;
        color: $text;
        margin-bottom: 1;
        margin-top: 1;
        background: $surface;
    }

    #help-grid {
        layout: grid;
        grid-size: 2;
        grid-columns: 1fr 1fr;
        grid-gutter: 0 2;
        width: 100%;
        height: auto;
        margin-bottom: 2;
        align: center middle;
    }

    .key {
        text-align: left;
        text-style: bold;
        color: $accent;
    }

    .desc {
        text-align: right;
        color: $text;
    }

    #help-exit {
        width: 100%;
        text-align: center;
        color: $text-muted;
        margin-bottom: 1;
    }
    """

    KEYBINDINGS = [
        ("Arrows", "Navigate tree"),
        ("/", "Search "),
        ("n / N", "Next / Prev match"),
        (":", "Jump to path"),
        ("c", "Copy path"),
        ("y", "Copy source"),
        ("t", "Toggle theme"),
        ("q", "Quit"),
    ]

    def compose(self) -> ComposeResult:
        try:
            version = importlib.metadata.version("twg")
        except importlib.metadata.PackageNotFoundError:
            version = "unknown"

        with Container(id="help-container"):
            # Branding Section
            yield Label(self.LOGO, id="help-logo")
            yield Label("Inspect. Navigate. Understand.", classes="help-tagline")
            yield Label(f"v{version}", classes="help-version")
            yield Label(f"v{version}", classes="help-version")
            yield Label("https://twig.wtf", id="help-link")
            yield Label("☕ buymeacoffee.com/workdone0", classes="help-tagline", id="coffee-link")
            
            # Keybindings Section
            yield Label("Keyboard Shortcuts", id="help-title")
            
            with Grid(id="help-grid"):
                for key, desc in self.KEYBINDINGS:
                    yield Label(key, classes="key")
                    yield Label(desc, classes="desc")

            yield Label("(Press Esc to close)", id="help-exit")

    def key_escape(self) -> None:
        self.dismiss(None)
    



from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Static, Label
from textual.reactive import reactive
import os
from twg.core.model import Node, DataType

class StatusBar(Horizontal):
    """
    A persistent status bar displaying file info and selection context.
    """
    
    selected_node: reactive[Node | None] = reactive(None)
    search_stats: reactive[str | None] = reactive(None)

    def __init__(self, file_path: str, model=None, **kwargs):
        self.file_path = file_path
        self.model = model
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        # File Info (Left)
        try:
            size_bytes = os.path.getsize(self.file_path)
            if size_bytes < 1024:
                size_str = f"{size_bytes} B"
            elif size_bytes < 1024 * 1024:
                size_str = f"{size_bytes / 1024:.1f} KB"
            else:
                size_str = f"{size_bytes / (1024 * 1024):.1f} MB"
        except OSError:
            size_str = "?"

        filename = os.path.basename(self.file_path)
        yield Static(f"FILE: {filename} ({size_str})", id="sb-file")
        
        yield Static("", id="sb-context")
        
        yield Static("READ ONLY", id="sb-mode")
        
        yield Static("", id="sb-search-stats")

    def watch_selected_node(self, node: Node | None) -> None:
        context = self.query_one("#sb-context", Static)
        
        if node is None:
            context.update("")
            return
            
        # Format: "Key (Type)"
        key = node.key if node.key else "root"
        type_str = node.type.value.capitalize()
        
        info = f"{key} : {type_str}"
        
        if node.is_container and self.model:
            count = self.model.get_children_count(node.id)
            info += f" [{count} items]"
                
        context.update(info)
        
    def watch_search_stats(self, stats: str | None) -> None:
        stats_widget = self.query_one("#sb-search-stats", Static)
        if stats:
            stats_widget.update(f"SEARCH: {stats}")
        else:
            stats_widget.update("")




import unittest
import json
from twg.core.cleaner import repair_json

class TestCleaner(unittest.TestCase):
    def test_repair_json_valid(self):
        valid = '{"a": 1}'
        self.assertEqual(json.loads(repair_json(valid)), {"a": 1})

    def test_repair_json_trailing_comma(self):
        content = '{"a": 1,}'
        self.assertEqual(json.loads(repair_json(content)), {"a": 1})

    def test_repair_json_single_quotes(self):
        content = "{'a': 1}"
        self.assertEqual(json.loads(repair_json(content)), {"a": 1})

    def test_repair_json_unquoted_keys(self):
        content = '{a: 1}'
        self.assertEqual(json.loads(repair_json(content)), {"a": 1})

    def test_repair_json_missing_brace(self):
        content = '{"a": 1'
        self.assertEqual(json.loads(repair_json(content)), {"a": 1})

    def test_repair_json_missing_bracket_and_brace(self):
        content = '{"a": [1, 2'
        self.assertEqual(json.loads(repair_json(content)), {"a": [1, 2]})

    def test_repair_json_mixed_errors(self):
        # Trailing comma, single quotes, unquoted key, missing brace
        content = "{'a': [1, 2, ], b: 3"
        expected = {"a": [1, 2], "b": 3}
        self.assertEqual(json.loads(repair_json(content)), expected)

if __name__ == '__main__':
    unittest.main()



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Input, Label
from textual.containers import Vertical, Container


class SearchModal(ModalScreen[str]):
    """Modal screen for entering search query."""

    CSS = """
    SearchModal {
        align: center middle;
    }

    #search-container {
        width: 60;
        height: auto;
        background: $surface;
        border: thick $primary;
        padding: 1 2;
        align: center middle;
    }

    #search-label {
        width: 100%;
        text-align: center;
        margin-bottom: 1;
        text-style: bold;
    }

    #search-input {
        margin-bottom: 1;
    }

    #search-help {
        width: 100%;
        text-align: center;
        color: $text-muted;
    }
    """

    def compose(self) -> ComposeResult:
        with Container(id="search-container"):
            yield Label("Search keys & values:", id="search-label")
            yield Input(placeholder="Type to search...", id="search-input")
            yield Label("(Press Esc to cancel)", id="search-help")

    def on_input_submitted(self, message: Input.Submitted) -> None:
        if message.value:
            self.dismiss(message.value)
        else:
            self.dismiss(None)
            
    def key_escape(self) -> None:
        self.dismiss(None)




import json
import json_repair

def repair_json(content: str) -> str:
    """
    Attempts to repair malformed JSON content using json-repair.
    
    Handles:
    - Trailing commas
    - Missing quotes
    - Unquoted keys
    - Missing colons
    - Comments
    - Unexpected characters
    """
    if not content.strip():
        return content

    try:
        # Load the repaired object
        obj = json_repair.repair_json(content, return_objects=True)
        
        # Sanitize object (replace NaN/Infinity with None)
        def sanitize(o):
            if isinstance(o, float):
                if o != o: # NaN
                    return None
                if o == float('inf') or o == float('-inf'):
                    return None
                return o
            elif isinstance(o, dict):
                return {k: sanitize(v) for k, v in o.items()}
            elif isinstance(o, list):
                return [sanitize(v) for v in o]
            return o
            
        obj = sanitize(obj)
        
        # Dump it back nicely formatted
        return json.dumps(obj, indent=2)
    except Exception as e:
        raise ValueError(f"Failed to repair JSON: {e}")




from textual.theme import Theme

SOLARIZED_DARK = Theme(
    name="solarized-dark",
    primary="#268bd2",
    secondary="#2aa198",
    accent="#dc322f",
    foreground="#839496",
    background="#002b36",
    success="#859900",
    warning="#b58900",
    error="#dc322f",
    surface="#073642",
    panel="#073642",
    dark=True,
)

CATPPUCCIN_MOCHA = Theme(
    name="catppuccin-mocha",
    primary="#cba6f7",
    secondary="#89b4fa",
    accent="#f38ba8",
    foreground="#cdd6f4",
    background="#1e1e2e",
    success="#a6e3a1",
    warning="#f9e2af",
    error="#f38ba8",
    surface="#313244",
    panel="#313244",
    dark=True,
)



import ijson
import uuid
import sqlite3
import os
from pathlib import Path
from typing import Callable, Optional
from twg.core.model import DataType
from twg.adapters.base_loader import BaseLoader, ProgressFile

class SQLiteLoader(BaseLoader):
    """
    Loads JSON file into SQLite database using ijson streaming.
    """
    
    def __init__(self):
        super().__init__()

    def ingest_file(self, file_path: str, db_path: Path, progress_callback: Optional[Callable[[int, int], None]] = None) -> None:
        """
        Parses JSON and inserts into DB.
        """
        conn = self.db_manager.get_connection(db_path)
        
        try:
            # High-performance bulk load settings
            conn.execute("PRAGMA synchronous = OFF")
            conn.execute("PRAGMA journal_mode = MEMORY")
            
            self._drop_indexes(conn)
            conn.commit()
            
            with open(file_path, 'rb') as f:
                self._parse_and_insert(f, conn)
            
            conn.commit()
            
            self._rebuild_indexes(conn)
            conn.commit()
            
        except Exception as e:
             raise e
        finally:
            conn.close()

    def _drop_indexes(self, conn: sqlite3.Connection):
        """Drop indexes and FTS table to speed up insertion."""
        conn.executescript("""
            DROP INDEX IF EXISTS idx_parent_rank;
            DROP INDEX IF EXISTS idx_path;
            DROP TRIGGER IF EXISTS nodes_ai;
            DROP TRIGGER IF EXISTS nodes_ad;
            DROP TRIGGER IF EXISTS nodes_au;
            DROP TABLE IF EXISTS nodes_search;
        """)

    def _rebuild_indexes(self, conn: sqlite3.Connection):
        """Re-create indexes and FTS table after bulk insert."""
        conn.executescript("""
            -- Standard Indexes
            CREATE INDEX IF NOT EXISTS idx_parent_rank ON nodes(parent_id, rank);
            CREATE INDEX IF NOT EXISTS idx_path ON nodes(path);

            -- FTS5 Search Table
            CREATE VIRTUAL TABLE IF NOT EXISTS nodes_search USING fts5(
                key, 
                value, 
                path, 
                content='nodes', 
                content_rowid='rowid'
            );

            -- Populate FTS
            INSERT INTO nodes_search(rowid, key, value, path) 
            SELECT rowid, key, value, path FROM nodes;

            -- Triggers for updates
            CREATE TRIGGER IF NOT EXISTS nodes_ai AFTER INSERT ON nodes BEGIN
              INSERT INTO nodes_search(rowid, key, value, path) VALUES (new.rowid, new.key, new.value, new.path);
            END;

            CREATE TRIGGER IF NOT EXISTS nodes_ad AFTER DELETE ON nodes BEGIN
              INSERT INTO nodes_search(nodes_search, rowid, key, value, path) VALUES('delete', old.rowid, old.key, old.value, old.path);
            END;

            CREATE TRIGGER IF NOT EXISTS nodes_au AFTER UPDATE ON nodes BEGIN
              INSERT INTO nodes_search(nodes_search, rowid, key, value, path) VALUES('delete', old.rowid, old.key, old.value, old.path);
              INSERT INTO nodes_search(rowid, key, value, path) VALUES (new.rowid, new.key, new.value, new.path);
            END;
        """)

    def _parse_and_insert(self, file_obj, conn: sqlite3.Connection):
        """
        Uses ijson to stream events and build the tree.
        """
        # ijson parser
        parser = ijson.parse(file_obj)
        
        # Helper to generate UUIDs
        def new_id(): return str(uuid.uuid4())
        
        root_id = new_id()
        
        # Stack items: (node_id, type, count/index, path_prefix)
        stack = [] 
        
        # Generator for batch inserts
        # We need to collect rows and executemany
        batch_size = 10000
        rows = []
        
        cursor = conn.cursor()
        
        def flush():
            if rows:
                cursor.executemany(
                    "INSERT INTO nodes (id, parent_id, key, value, type, rank, path, is_expanded) VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                    rows
                )
                rows.clear()
        
        # ijson events stream
        first_event = True
        
        current_key = None # For map values
        
        for prefix, event, value in parser:
            
            if first_event:
                first_event = False
                # Determine root type
                root_type = DataType.NULL
                if event == 'start_map':
                    root_type = DataType.OBJECT
                elif event == 'start_array':
                    root_type = DataType.ARRAY
                else: 
                     # Primitive root
                     root_type = DataType.from_value(value)
                     
                root_row = (
                    root_id, 
                    None, 
                    "root", 
                    str(value) if not (event == 'start_map' or event == 'start_array') else None,
                    root_type.value,
                    0,
                    ".", # Root path
                    0
                )
                rows.append(root_row)
                
                if event in ('start_map', 'start_array'):
                    stack.append([root_id, root_type, 0, "."])
                continue
            
            # Processing children
            if not stack:
                 break
                 
            parent_id, parent_type, rank, parent_path = stack[-1]
            
            if event == 'end_map' or event == 'end_array':
                stack.pop()
                continue
            
            if event == 'map_key':
                current_key = value
                continue
            
            # Now we have a value (container start or primitive)
            
            # Determine Key
            if parent_type == DataType.ARRAY:
                node_key = str(rank)
            else:
                node_key = current_key
            
            # Determine Path
            if parent_type == DataType.ARRAY:
                node_path = f"{parent_path}[{node_key}]"
            else:
                if parent_path == ".":
                    node_path = f".{node_key}"
                else:
                    node_path = f"{parent_path}.{node_key}"
            
            # Update rank for parent
            stack[-1][2] += 1
            
            # Determine Type
            node_type = DataType.STRING # Default
            val_str = None
            
            is_container = False
            
            if event == 'start_map':
                node_type = DataType.OBJECT
                is_container = True
            elif event == 'start_array':
                node_type = DataType.ARRAY
                is_container = True
            elif event == 'boolean':
                node_type = DataType.BOOLEAN
                val_str = str(value)
            elif event == 'integer':
                node_type = DataType.INTEGER
                val_str = str(value)
            elif event == 'double': # ijson uses 'double' for floats
                node_type = DataType.FLOAT
                val_str = str(value)
            elif event == 'number': 
                node_type = DataType.FLOAT # generic number
                val_str = str(value)
            elif event == 'string':
                node_type = DataType.STRING
                val_str = value
            elif event == 'null':
                node_type = DataType.NULL
                val_str = "null"
            
            node_id = new_id()
            
            rows.append((
                node_id,
                parent_id,
                node_key,
                val_str,
                node_type.value,
                rank,
                node_path,
                0
            ))
            
            if len(rows) >= batch_size:
                flush()
                
            if is_container:
                stack.append([node_id, node_type, 0, node_path])
        
        flush()



import os
import json
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch
from twg.core.config import Config

class TestConfig(unittest.TestCase):
    def setUp(self):
        self.temp_dir = TemporaryDirectory()
        self.config_dir = Path(self.temp_dir.name) / "config"
        
        # Patch the _get_config_dir method to return our temp dir
        self.patcher = patch.object(Config, '_get_config_dir', return_value=self.config_dir)
        self.mock_get_dir = self.patcher.start()

    def tearDown(self):
        self.patcher.stop()
        self.temp_dir.cleanup()

    def test_default_config(self):
        config = Config()
        self.assertEqual(config.get("theme"), "catppuccin-mocha")

    def test_save_and_load(self):
        config = Config()
        config.set("theme", "dracula")
        
        # Create a new instance to verify persistence
        new_config = Config()
        self.assertEqual(new_config.get("theme"), "dracula")
        
        # Verify file content
        config_file = self.config_dir / "config.json"
        self.assertTrue(config_file.exists())
        with open(config_file, "r") as f:
            data = json.load(f)
            self.assertEqual(data["theme"], "dracula")

    def test_load_existing_file(self):
        # Create a config file manually
        self.config_dir.mkdir(parents=True)
        config_file = self.config_dir / "config.json"
        with open(config_file, "w") as f:
            json.dump({"theme": "solarized-light"}, f)
            
        config = Config()
        self.assertEqual(config.get("theme"), "solarized-light")

if __name__ == '__main__':
    unittest.main()



from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, LoadingIndicator, Label, Static
from textual.containers import Container, Horizontal, Vertical, Center, Middle
import sys
import os
import argparse
import pyperclip
from rich.console import Console
from rich.json import JSON
from rich.panel import Panel
from rich.text import Text
from rich import box
from rich.syntax import Syntax

from textual.binding import Binding

try:
    import yaml
except ImportError:
    yaml = None

from twg.core.model import SQLiteModel
from twg.core.config import Config
from twg.adapters.sqlite_loader import SQLiteLoader
from twg.adapters.yaml_loader import YamlLoader
from twg.ui.widgets.navigator import ColumnNavigator
from twg.ui.widgets.inspector import Inspector
from twg.ui.widgets.status_bar import StatusBar
from twg.ui.widgets.search import SearchModal
from twg.ui.widgets.jump import JumpModal
from twg.ui.widgets.help import HelpModal
from twg.ui.widgets.loading import LoadingScreen
from twg.ui.widgets.breadcrumbs import Breadcrumbs
from twg.core.model import Node
from twg.core.cleaner import repair_json

class TwigApp(App):
    """
    The main Textual application class for Twig.
    
    Manages the application lifecycle, layout composition, and global actions
    like theme toggling and path copying.
    """
    CSS_PATH = "styles.tcss"

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("c", "copy_path", "Copy Path"),
        ("y", "copy_source", "Copy Source"),
        ("t", "toggle_theme", "Toggle Theme"),
        ("/", "search", "Search"),
        ("n", "next_match", "Next Match"),
        ("N", "prev_match", "Prev Match"),
        (":", "jump", "Jump to Path"),
        ("?", "help", "Help"),
        Binding("h", "help", "Help", show=False),
    ]

    def on_mount(self) -> None:
        from twg.ui.themes import SOLARIZED_DARK, CATPPUCCIN_MOCHA
        self.register_theme(SOLARIZED_DARK)
        self.register_theme(CATPPUCCIN_MOCHA)
        
        self.config = Config()
        self.theme = self.config.get("theme", "catppuccin-mocha")
        self.title = "Twig"
        self.run_worker(self.load_file, thread=True)

    def load_file(self) -> None:
        """Worker method to load the file in a background thread."""
        try:
            if self.file_path.lower().endswith(('.yaml', '.yml')):
                self.format = "yaml"
                loader = YamlLoader()
            else:
                self.format = "json"
                loader = SQLiteLoader()
            
            model = loader.load_into_model(
                self.file_path, 
                force_rebuild=self.force_rebuild
            )
            self.call_from_thread(self.on_file_loaded, model)
        except Exception as e:
            self.call_from_thread(self.on_file_load_error, str(e))
    
    def on_file_loaded(self, model: SQLiteModel) -> None:
        """Called when the file is successfully loaded."""
        self.model = model
        
        # Dismiss loading screen if it exists
        try:
            self.pop_screen()
        except:
            pass
            
        content = self.query_one("#main-content")
        content.remove_children()
        
        content.mount(
            Vertical(
                Breadcrumbs(self.model, id="breadcrumbs"),
                Horizontal(
                    ColumnNavigator(self.model),
                    Inspector(self.model, format=self.format, id="inspector"),
                    id="main-container"
                ),
                id="content-view"
            )
        )
        
        status_bar = self.query_one(StatusBar)
        status_bar.model = self.model

    def on_file_load_error(self, error_message: str) -> None:
        """Called when file loading fails."""
        # Exit the app and pass the error message back to the caller
        
        # Colorized hint using Rich markup
        
        hint = (
            f"\n\n[bold yellow]Tip:[/bold yellow] If the file contains invalid JSON, try running:\n"
            f"  [green]twg --fix {self.file_path} -o {self.file_path}[/green]\n"
            f"[dim](This will repair and overwrite the file in place)[/dim]"
        )
        self.exit(f"[bold red]Error:[/bold red] {error_message}{hint}")

    def action_toggle_theme(self) -> None:
        """Cycles through all available themes and notifies the user."""
        themes = [t for t in self.available_themes.keys() if t != "textual-ansi"]
        try:
            current_index = themes.index(self.theme)
        except ValueError:
            current_index = 0
            
        next_index = (current_index + 1) % len(themes)
        next_theme = themes[next_index]
        
        self.theme = next_theme
        self.config.set("theme", next_theme)
        self.notify(f"Theme: {next_theme}")

    def __init__(self, file_path: str, force_rebuild: bool = False):
        self.file_path = file_path
        self.force_rebuild = force_rebuild
        self.model: SQLiteModel | None = None
        self.current_node: Node | None = None
        self.last_search_query: str | None = None
        self.config: Config | None = None
        self.format: str = "json" # Default
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Header(show_clock=False)
        
        with Container(id="main-content"):
            yield Label("Initializing...", id="init-label")
            
        yield StatusBar(self.file_path, id="status-bar")
        yield Footer()

    def on_ready(self) -> None:
        """Push loading screen once app is ready."""
        self.push_screen(LoadingScreen(message=f"Loading {os.path.basename(self.file_path)}..."))

    def on_column_navigator_node_selected(self, message: ColumnNavigator.NodeSelected) -> None:
        """
        Handler for node selection events from the ColumnNavigator.
        
        Updates the Inspector, Breadcrumbs, and Status Bar with the selected node details.
        """
        node = self.model.get_node(message.node_id)
        self.current_node = node
        
        inspector = self.query_one(Inspector)
        inspector.selected_node = node
        
        breadcrumbs = self.query_one(Breadcrumbs)
        breadcrumbs.selected_node = node
        
        status_bar = self.query_one(StatusBar)
        status_bar.selected_node = node

    def action_search(self) -> None:
        """Open the search modal."""
        async def check_search(query: str | None) -> None:
            if query:
                self.last_search_query = query
                await self.action_next_match()
        
        self.push_screen(SearchModal(), check_search)
            
    def action_help(self) -> None:
        """Open the help modal."""
        self.push_screen(HelpModal())

    async def action_jump(self) -> None:
        """Open the jump modal."""
        async def check_jump(path: str | None) -> None:
            if path:
                if not self.model: return
                
                try:
                    node = self.model.resolve_path(path)
                    if node:
                        navigator = self.query_one(ColumnNavigator)
                        await navigator.expand_to_node(node.id)
                    else:
                        self.notify(f"Path not found: {path}", severity="error")
                except ValueError as e:
                    self.notify(str(e), severity="error")

        await self.push_screen(JumpModal(), check_jump)

    async def action_next_match(self) -> None:
        """Find next match for the last query."""
        if not self.last_search_query:
            self.notify("No active search query.", severity="warning")
            return
            
        navigator = self.query_one(ColumnNavigator)
        
        found_node = await navigator.find_next(self.last_search_query, direction=1)
        if not found_node:
             self.notify(f"Not found: '{self.last_search_query}'", severity="warning")
        else:
             # Show stats (current/total)
             if self.model:
                 current, total = self.model.get_search_stats(self.last_search_query, found_node.id)
                 if total > 0:
                     self.query_one(StatusBar).search_stats = f"{current}/{total}"

    async def action_prev_match(self) -> None:
        """Find previous match for the last search query."""
        if not self.last_search_query:
            self.notify("No active search query.", severity="warning")
            return

        navigator = self.query_one(ColumnNavigator)
        
        found_node = await navigator.find_next(self.last_search_query, direction=-1)
        if not found_node:
             self.notify(f"Not found: '{self.last_search_query}'", severity="warning")
        else:
             # Show stats
             if self.model:
                 current, total = self.model.get_search_stats(self.last_search_query, found_node.id)
                 if total > 0:
                     self.query_one(StatusBar).search_stats = f"{current}/{total}"

    def action_copy_path(self) -> None:
        """Copies the jq-style path of the currently selected node to the clipboard."""
        if not self.current_node:
            return
            
        full_path = self.model.get_path(self.current_node.id)
        if full_path:
            try:
                pyperclip.copy(full_path)
                self.notify(f"Copied path: {full_path}")
            except Exception as e:
                self.notify(f"Failed to copy: {e}", severity="error")
        else:
             self.notify("Root path copied")

    def action_copy_source(self) -> None:
        """Copies the raw source (JSON/YAML) of the selected node to clipboard."""
        if not self.current_node:
            return
            
        import json
        
        try:
            if self.current_node.is_container:
                data = self.model.reconstruct_json(self.current_node.id, max_depth=5)
            else:
                data = self.current_node.value

            if self.format == "yaml" and yaml:
                source_str = yaml.dump(data, sort_keys=False, default_flow_style=False)
                label = "YAML source"
            else:
                source_str = json.dumps(data, indent=2)
                label = "source"
            
            pyperclip.copy(source_str)
            
            preview = source_str[:50].replace('\n', ' ')
            if len(source_str) > 50: preview += "..."
            
            self.notify(f"Copied {label}: {preview}")
            
        except Exception as e:
            self.notify(f"Failed to copy source: {e}", severity="error")

def run():
    parser = argparse.ArgumentParser(
        description="Twig: Inspect. Navigate. Understand. A modern, terminal-based data explorer."
    )
    parser.add_argument(
        "file",
        help="The JSON/data file to explore."
    )
    parser.add_argument(
        "-o", "--output",
        help="Output file. For --fix, it saves the repaired JSON. For --print, it saves the formatted JSON. If omitted, prints to stdout."
    )
    
    try:
        from importlib.metadata import version, PackageNotFoundError
        twg_version = version("twg")
    except ImportError:
        twg_version = "unknown"
    except PackageNotFoundError:
        twg_version = "dev"

    parser.add_argument(
        "-v", "--version",
        action="version",
        version=f"%(prog)s {twg_version}"
    )
    parser.add_argument(
        "--fix",
        action="store_true",
        help="Attempt to automatically repair malformed JSON and exit."
    )
    parser.add_argument(
        "-p", "--print",
        action="store_true",
        help="Print formatted and syntax-highlighted JSON to stdout and exit."
    )
    
    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help="Number of spaces for indentation (default: 2)."
    )
    parser.add_argument(
        "--rebuild-db",
        action="store_true",
        help="Force rebuild of the internal SQLite database cache."
    )
    
    args = parser.parse_args()

    if not args.fix and not args.print:
        if not os.path.exists(args.file):
            print(f"Error: File not found: {args.file}", file=sys.stderr)
            sys.exit(1)

        if not args.file.lower().endswith((".json", ".yaml", ".yml")):
            print(f"Error: Invalid file type '{args.file}'. Twig currently only supports .json and .yaml files.", file=sys.stderr)
            sys.exit(1)
            
        app = TwigApp(args.file, force_rebuild=args.rebuild_db)
        result = app.run()
        
        if result is not None and isinstance(result, str):
            console = Console(stderr=True)
            console.print(result)
            sys.exit(1)
        return

    if not os.path.exists(args.file):
        print(f"Error: File not found: {args.file}", file=sys.stderr)
        sys.exit(1)

    is_yaml = args.file.lower().endswith(('.yaml', '.yml'))

    if is_yaml:
        if args.fix:
            print("Error: Fix/Repair is currently only supported for JSON files.", file=sys.stderr)
            sys.exit(1)
        
        try:
            with open(args.file, 'r', encoding='utf-8') as f:
                content = f.read()
            
            try:
                if yaml:
                    parsed = yaml.safe_load(content)
                else:
                    print("Error: PyYAML is not installed. Cannot process YAML files.", file=sys.stderr)
                    sys.exit(1)
            except Exception as e:
                print(f"Error: Failed to parse YAML: {e}", file=sys.stderr)
                sys.exit(1)
                
            if args.output:
                with open(args.output, 'w', encoding='utf-8') as f:
                    yaml.dump(parsed, f, sort_keys=False, default_flow_style=False)
                print(f"Formatted YAML written to {args.output}", file=sys.stderr)
            
            else:
                file_size = len(content.encode('utf-8'))
                
                if isinstance(parsed, list):
                    item_count = len(parsed)
                    type_label = "Array"
                elif isinstance(parsed, dict):
                    item_count = len(parsed)
                    type_label = "Object"
                else:
                    item_count = 1
                    type_label = "Primitive"
                
                err_console = Console(stderr=True)
                meta_text = Text()
                meta_text.append(f" File: ", style="bold cyan")
                meta_text.append(f"{os.path.basename(args.file)} \n")
                meta_text.append(f" Size: ", style="bold cyan")
                meta_text.append(f"{file_size / 1024:.1f} KB \n")
                meta_text.append(f" Type: ", style="bold cyan")
                meta_text.append(f"{type_label} ({item_count} items)")
                
                err_console.print(Panel(
                    meta_text,
                    title="Twig Metadata",
                    border_style="blue",
                    box=box.ROUNDED,
                    expand=False
                ))

                console = Console()
                syntax = Syntax(content, "yaml", theme="monokai", line_numbers=False)
                console.print(syntax)
                
        except Exception as e:
            print(f"Error processing YAML file: {e}", file=sys.stderr)
            sys.exit(1)
        
        sys.exit(0)

    try:
        with open(args.file, 'r', encoding='utf-8') as f:
            content = f.read()

        if args.fix:
            try:
                content = repair_json(content)
            except Exception as e:
                print(f"Error repairing JSON: {e}", file=sys.stderr)
                sys.exit(1)
        
        try:
            import json
            parsed = json.loads(content)
        except json.JSONDecodeError as e:
            if args.print:
                 print(f"Error: Failed to parse JSON. Use --fix to attempt repair.\n{e}", file=sys.stderr)
                 sys.exit(1)
            else:
                 print(f"Error: Resulting JSON is invalid even after attempted fix.\n{e}", file=sys.stderr)
                 sys.exit(1)

        if args.output:
            # Write to file (formatted)
            with open(args.output, 'w', encoding='utf-8') as f:
                json.dump(parsed, f, indent=args.indent)
            
            action = "Fixed" if args.fix else "Formatted"
            print(f"{action} JSON written to {args.output}", file=sys.stderr)
        
        else:
            if args.print:
                file_size = len(content.encode('utf-8'))
                
                if isinstance(parsed, list):
                    item_count = len(parsed)
                    type_label = "Array"
                elif isinstance(parsed, dict):
                    item_count = len(parsed)
                    type_label = "Object"
                else:
                    item_count = 1
                    type_label = "Primitive"
                
                err_console = Console(stderr=True)
                meta_text = Text()
                meta_text.append(f" File: ", style="bold cyan")
                meta_text.append(f"{os.path.basename(args.file)} \n")
                meta_text.append(f" Size: ", style="bold cyan")
                meta_text.append(f"{file_size / 1024:.1f} KB \n")
                meta_text.append(f" Type: ", style="bold cyan")
                meta_text.append(f"{type_label} ({item_count} items)")
                
                err_console.print(Panel(
                    meta_text,
                    title="Twig Metadata",
                    border_style="blue",
                    box=box.ROUNDED,
                    expand=False
                ))

                # JSON to stdout
                console = Console()
                console.print(JSON(content, indent=args.indent))
            
            else:
                print(json.dumps(parsed, indent=args.indent))

        sys.exit(0)

    except Exception as e:
        print(f"Error processing file: {e}", file=sys.stderr)
        sys.exit(1)
    

if __name__ == "__main__":
    run()



import asyncio
import unittest
import os
from textual.pilot import Pilot
from twg.ui.app import TwigApp

# Sample Data Path
SAMPLE_PATH = "samples/cloud_infrastructure.json"

class TestTwigIntegration(unittest.IsolatedAsyncioTestCase):
    """
    Comprehensive Integration Test Suite for Twig.
    Covers: Jump, Navigation, Global Search, Smart Search, Directional Search, and Themes.
    """

    async def asyncSetUp(self):
        if not os.path.exists(SAMPLE_PATH):
            self.fail(f"Sample file {SAMPLE_PATH} not found.")
        self.app = TwigApp(SAMPLE_PATH)

    async def test_full_user_journey(self):
        """
        Simulates a complete user session exercising core features.
        Runs as a single flow to mimic real usage and state accumulation.
        """
        async with self.app.run_test() as pilot:
            await pilot.pause(0.5)
            navigator = self.app.query_one("ColumnNavigator")

            # ==========================================
            # 1. SMART SEARCH (Path Resolution)
            # ==========================================
            print("\n[TEST] 1. Smart Search (Path Jump)...")
            # Search for a specific deep path using the Search Bar
            # Path: .regions["us-east-1"].vpcs[0].subnets[1].instances[0].name -> "api-gateway"
            target_path = '.regions["us-east-1"].vpcs[0].subnets[1].instances[0].name'
            
            await pilot.press("/")
            await pilot.press(*target_path)
            await pilot.press("enter")
            
            # Wait for expansion (with buffer for asyncio.sleep in navigator)
            await pilot.pause(0.8) 
            
            # Verify focus
            focused = self.app.screen.focused
            print(f"       Focused: {focused}")
            # Should be in a deep column (col-6 typically for this depth: regions->us-east-1->vpcs->0->subnets->1->instances->0->name)
            self.assertIn("col-", str(focused.id), "Smart Search should jump to column")
            
            # ==========================================
            # 2. NAVIGATION (Miller Columns)
            # ==========================================
            print("[TEST] 2. Navigation (Left/Right)...")
            # Move Left (Parent)
            await pilot.press("left")
            await pilot.pause(0.2)
            focused_left = self.app.screen.focused
            print(f"       Left Focus: {focused_left}")
            
            # Move Right (Child) - should return to previous selection
            await pilot.press("right")
            await pilot.pause(0.2)
            focused_right = self.app.screen.focused
            print(f"       Right Focus: {focused_right}")

            # ==========================================
            # 3. GLOBAL SEARCH & DIRECTION (Next/Prev)
            # ==========================================
            print("[TEST] 3. Global Search & Direction...")
            # Focus root to reset context
            col0 = self.app.query_one("#col-0")
            col0.focus()
            
            # Search for "available" - this appears multiple times (DBs)
            await pilot.press("/")
            await pilot.press(*"available")
            await pilot.press("enter")
            await pilot.pause(0.5)
            
            # Capture first match ID
            col_match_1 = self._get_focused_column(navigator)
            node_id_1 = self._get_selected_node_id(col_match_1)
            print(f"       Match 1 ID: {node_id_1}")
            
            # Find Next ('n')
            await pilot.press("n")
            await pilot.pause(0.5)
            
            col_match_2 = self._get_focused_column(navigator)
            node_id_2 = self._get_selected_node_id(col_match_2)
            print(f"       Match 2 ID: {node_id_2}")
            
            self.assertNotEqual(node_id_1, node_id_2, "Next (n) should move to a different node")
            
            # Find Previous ('N') - Should return to Match 1 (wrapping or direct prev)
            # Note: Depending on count, prev might be match 1.
            await pilot.press("N")
            await pilot.pause(0.5)
            
            col_match_prev = self._get_focused_column(navigator)
            node_id_prev = self._get_selected_node_id(col_match_prev)
            print(f"       Prev Match ID: {node_id_prev}")
            
            self.assertEqual(node_id_prev, node_id_1, "Prev (N) should return to previous match")

            # ==========================================
            # 4. THEME CYCLING
            # ==========================================
            print("[TEST] 4. Theme Cycling...")
            initial_theme = self.app.theme
            
            await pilot.press("t")
            new_theme = self.app.theme
            print(f"       Theme: {initial_theme} -> {new_theme}")
            
            self.assertNotEqual(initial_theme, new_theme, "Theme should change on pressing 't'")

            print("\n[SUCCESS] Integration Test Suite Passed.")

    def _get_focused_column(self, navigator):
        for child in navigator.children:
            if "Column" in str(type(child)) and child.query_one("TwigOptionList").has_focus:
                return child
        return None

    def _get_selected_node_id(self, column):
        if not column: return None
        opts = column.query_one("TwigOptionList")
        if opts.highlighted is not None:
             return column.node_map.get(opts.highlighted)
        return None

if __name__ == "__main__":
    unittest.main()



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Label, LoadingIndicator
from textual.containers import Container, Center, Middle, Vertical


class LoadingScreen(ModalScreen):
    """A modal screen that shows a loading indicator."""

    LOGO = r"""
████████╗██╗    ██╗██╗ ██████╗ 
╚══██╔══╝██║    ██║██║██╔════╝ 
   ██║   ██║ █╗ ██║██║██║  ███╗
   ██║   ██║███╗██║██║██║   ██║
   ██║   ╚███╔███╔╝██║╚██████╔╝
   ╚═╝    ╚══╝╚══╝ ╚═╝ ╚═════╝ 
"""

    CSS = """
    LoadingScreen {
        align: center middle;
    }

    #loading-container {
        width: 60;
        height: auto;
        background: $surface;
        border: thick $primary;
        padding: 1 2;
    }
    
    #loading-logo {
        width: 100%;
        text-align: center;
        color: $warning;
        margin-bottom: 1;
    }

    #loading-bar {
        width: 100%;
        margin-bottom: 1;
    }

    #loading-label {
        width: 100%;
        text-align: center;
        text-style: bold;
    }

    .loading-link {
        width: 100%;
        text-align: center;
        color: $text-muted;
        margin-top: 1;
    }
    """

    def __init__(self, message: str = "Working...", name: str | None = None, id: str | None = None, classes: str | None = None):
        self.message = message
        super().__init__(name, id, classes)

    def compose(self) -> ComposeResult:
        with Container(id="loading-container"):
            yield Label(self.LOGO, id="loading-logo")
            yield Center(LoadingIndicator())
            yield Label(self.message, id="loading-label")
            yield Label("https://twig.wtf", classes="loading-link")
            yield Label("☕ buymeacoffee.com/workdone0", classes="loading-link")



import yaml
import uuid
import sqlite3
import os
from pathlib import Path
from typing import Any, Callable, Optional  
from twg.core.model import DataType
from twg.adapters.base_loader import BaseLoader



class YamlLoader(BaseLoader):
    """
    Loads YAML file into SQLite database using PyYAML event stream.
    """
    
    def ingest_file(self, file_path: str, db_path: Path, progress_callback: Optional[Callable[[int, int], None]] = None) -> None:
        conn = self.db_manager.get_connection(db_path)
        
        try:
            total_size = os.path.getsize(file_path)
            
            with open(file_path, 'rb') as f:
                if progress_callback:
                    wrapped_f = ProgressFile(f, total_size, progress_callback)
                    self._parse_and_insert(wrapped_f, conn)
                else:
                    self._parse_and_insert(f, conn)
            
            conn.commit()
            
            # Index
            conn.execute("INSERT INTO nodes_search(rowid, key, value, path) SELECT rowid, key, value, path FROM nodes")
            conn.commit()
            
        except Exception as e:
            raise e
        finally:
            conn.close()

    def _parse_and_insert(self, file_obj, conn: sqlite3.Connection):
        """
        Parses YAML events and inserts into DB.
        """
        # Use CLoader if available for speed
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader

        # Get events
        # yaml.parse returns a generator of events
        events = yaml.parse(file_obj, Loader=Loader)
        
        def new_id(): return str(uuid.uuid4())
        
        root_id = new_id()
        
        # Stack: [node_id, type, count, path_prefix]
        stack = [] 
        
        batch_size = 10000
        rows = []
        cursor = conn.cursor()
        
        def flush():
            if rows:
                cursor.executemany(
                    "INSERT INTO nodes (id, parent_id, key, value, type, rank, path, is_expanded) VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                    rows
                )
                rows.clear()
        
        current_key = None 
        in_document = False
        
        # Treat the stream as an array of documents (Virtual Root)
        virtual_root_type = DataType.ARRAY
        root_row = (root_id, None, "root", None, virtual_root_type.value, 0, ".", 0)
        rows.append(root_row)
        
        stack.append([root_id, virtual_root_type, 0, "."])
        
        for event in events:
            if isinstance(event, yaml.StreamStartEvent):
                continue
            if isinstance(event, yaml.StreamEndEvent):
                continue
                
            if isinstance(event, yaml.DocumentStartEvent):
                # Start of a new document
                continue
            
            if isinstance(event, yaml.DocumentEndEvent):
                continue
            
            # Now content events: MappingStart, SequenceStart, Scalar, Alias
            
            # Determine hierarchy context
            if not stack: break
            
            parent_id, parent_type, rank, parent_path = stack[-1]
            
            is_key_scalar = False
            
            if parent_type == DataType.OBJECT:
                if current_key is None:
                    # Expecting Key
                    if isinstance(event, yaml.ScalarEvent):
                        current_key = event.value
                        continue
                    elif isinstance(event, (yaml.MappingStartEvent, yaml.SequenceStartEvent)):
                        # Fallback for complex keys
                        current_key = f"<{type(event).__name__}>"
                        pass
                    elif isinstance(event, (yaml.MappingEndEvent, yaml.SequenceEndEvent)):
                         # End of map
                         stack.pop()
                         continue
                
            elif parent_type == DataType.ARRAY:
                 if isinstance(event, (yaml.SequenceEndEvent)):
                     stack.pop()
                     continue

            # Ensure we are processing a value
            if isinstance(event, (yaml.MappingEndEvent, yaml.SequenceEndEvent)):
                 stack.pop()
                 continue

            # Determine Node details
            if parent_type == DataType.ARRAY:
                 node_key = str(rank)
            else:
                 node_key = current_key
                 current_key = None # Reset for next pair
            
            # Path
            if parent_type == DataType.ARRAY:
                node_path = f"{parent_path}[{node_key}]"
            else:
                if parent_path == ".":
                    node_path = f".{node_key}"
                else:
                    node_path = f"{parent_path}.{node_key}"
            
            # Update parent rank
            stack[-1][2] += 1
            
            node_id = new_id()
            node_type = DataType.STRING
            val_str = None
            is_container = False
            
            if isinstance(event, yaml.ScalarEvent):
                # Detect type
                # PyYAML tags: event.tag (e.g. tag:yaml.org,2002:int)
                tag = event.tag
                val = event.value
                
                if tag and 'int' in tag:
                    node_type = DataType.INTEGER
                    val_str = str(val)
                elif tag and 'float' in tag:
                    node_type = DataType.FLOAT
                    val_str = str(val)
                elif tag and 'bool' in tag:
                    node_type = DataType.BOOLEAN
                    val_str = str(val).lower()
                elif tag and 'null' in tag:
                    node_type = DataType.NULL
                    val_str = "null"
                else:
                    # Infer if implicit
                    if event.style is None and event.implicit:
                        # minimalistic inference
                        if val.lower() in ('true', 'false'):
                            node_type = DataType.BOOLEAN
                            val_str = val.lower()
                        elif val.lower() in ('null', '~'):
                            node_type = DataType.NULL
                            val_str = "null"
                        elif val.isdigit():
                            node_type = DataType.INTEGER
                            val_str = val
                        else:
                            node_type = DataType.STRING
                            val_str = val
                    else:
                        node_type = DataType.STRING
                        val_str = val

            elif isinstance(event, yaml.MappingStartEvent):
                node_type = DataType.OBJECT
                is_container = True
            
            elif isinstance(event, yaml.SequenceStartEvent):
                node_type = DataType.ARRAY
                is_container = True
            
            elif isinstance(event, yaml.AliasEvent):
                node_type = DataType.STRING
                val_str = f"*alias:{event.anchor}"
            
            rows.append((
                node_id,
                parent_id,
                node_key,
                val_str,
                node_type.value,
                rank,
                node_path,
                0
            ))
            
            if len(rows) >= batch_size:
                flush()
            
            if is_container:
                stack.append([node_id, node_type, 0, node_path])

        flush()



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Input, Label
from textual.containers import Container

class JumpModal(ModalScreen[str]):
    """Modal screen for entering jq path to jump to."""

    CSS = """
    JumpModal {
        align: center middle;
    }

    #jump-container {
        width: 60;
        height: auto;
        background: $surface;
        border: thick $primary;
        padding: 1 2;
        align: center middle;
    }

    #jump-label {
        width: 100%;
        text-align: center;
        margin-bottom: 1;
        text-style: bold;
    }

    #jump-input {
        margin-bottom: 1;
    }

    #jump-exit {
        width: 100%;
        text-align: center;
        color: $text-muted;
        margin-top: 1;
    }
    """

    def compose(self) -> ComposeResult:
        with Container(id="jump-container"):
            yield Label("Jump to path (jq style):", id="jump-label")
            yield Input(placeholder=".items[0].id", id="jump-input")
            yield Label("(Press Esc to cancel)", id="jump-exit")

    def on_input_submitted(self, message: Input.Submitted) -> None:
        if message.value:
            self.dismiss(message.value)
        else:
            self.dismiss(None)
            
    def key_escape(self) -> None:
        self.dismiss(None)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Tuple, Callable
from twg.core.model import SQLiteModel
from twg.core.db import DatabaseManager

class ProgressFile:
    """Wrapper around a file object to track bytes read."""
    def __init__(self, file_obj, total_size: int, callback: Callable[[int, int], None]):
        self.file_obj = file_obj
        self.total_size = total_size
        self.callback = callback
        self.bytes_read = 0
        self._last_report = 0
        self._report_interval = 1024 * 1024  # Report every 1MB

    def read(self, size=-1):
        data = self.file_obj.read(size)
        if data:
            self.bytes_read += len(data)
            if self.bytes_read - self._last_report > self._report_interval or self.bytes_read >= self.total_size:
                 self.callback(self.bytes_read, self.total_size)
                 self._last_report = self.bytes_read
        return data

    def __getattr__(self, name):
        return getattr(self.file_obj, name)

class BaseLoader(ABC):
    """
    Abstract base class for file loaders (JSON, YAML, etc).
    """
    def __init__(self):
        self.db_manager = DatabaseManager()

    def prepare_db(self, file_path: str, force_rebuild: bool = False) -> Tuple[Path, bool]:
        """
        Common logic to prepare the DB file.
        Returns (db_path, needs_ingestion).
        """
        path = Path(file_path)
        if not path.exists():
             raise FileNotFoundError(f"File not found: {file_path}")
             
        db_path = self.db_manager.get_db_path(file_path)
        
        # Check if DB already exists
        if db_path.exists() and not force_rebuild:
            try:
                # Simple check if valid
                conn = self.db_manager.get_connection(db_path)
                cursor = conn.execute("SELECT count(*) FROM nodes")
                # verify it has data
                count = cursor.fetchone()[0]
                conn.close()
                if count > 0:
                    return db_path, False
            except:
                pass # Rebuild if check fails
        
        # Init new DB
        if db_path.exists():
            try:
                db_path.unlink()
            except OSError:
                pass
                 
        self.db_manager.init_db(db_path)
        return db_path, True

    @abstractmethod
    def ingest_file(self, file_path: str, db_path: Path, progress_callback: Callable[[int, int], None] | None = None) -> None:
        """
        Parse file and insert into DB. To be implemented by subclasses.
        """
        pass

    def load_into_model(self, file_path: str, force_rebuild: bool = False) -> SQLiteModel:
        """
        Synchronous helper.
        """
        db_path, needs_ingestion = self.prepare_db(file_path, force_rebuild)
        if needs_ingestion:
            self.ingest_file(file_path, db_path)
            
        return SQLiteModel(str(db_path))



from textual.app import ComposeResult
from textual.containers import Vertical, Container, Horizontal, VerticalScroll
from textual.widgets import Static, Label
from textual.reactive import reactive
from rich.text import Text
from rich.table import Table
from rich.panel import Panel
from rich.syntax import Syntax
from rich.json import JSON
from datetime import datetime
import json
import re

try:
    import yaml
except ImportError:
    yaml = None

from twg.core.model import Node, DataType, SQLiteModel

class Inspector(Container):
    """
    Displays detailed information about the selected node in a rich format.
    Uses a vertical stack layout for better use of vertical space.
    """
    
    selected_node: reactive[Node | None] = reactive(None)
    format: reactive[str] = reactive("json") # "json" or "yaml"
    
    # Regex patterns for smart insights
    URL_PATTERN = re.compile(r'https?://(?:[-\w.]|(?:%[\da-fA-F]{2}))+')
    HEX_COLOR_PATTERN = re.compile(r'^#(?:[0-9a-fA-F]{3}){1,2}$')
    EMAIL_PATTERN = re.compile(r'[^@]+@[^@]+\.[^@]+')

    SYNTAX_THEME = "monokai"

    def __init__(self, model: SQLiteModel, format: str = "json", **kwargs):
        self.model = model
        super().__init__(**kwargs)
        self.format = format

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("Inspector", id="inspector-title")
            yield Label("", id="insp-path")
            
            with VerticalScroll(id="inspector-scroll"):
                yield Container(id="insp-details-grid")
                yield Static(id="insp-insights", classes="insp-section")
                yield Static(id="insp-preview_content", classes="insp-section")
                yield Static(id="insp-raw_content", classes="insp-section")

    def watch_selected_node(self, node: Node | None) -> None:
        path_view = self.query_one("#insp-path", Label)
        details_grid = self.query_one("#insp-details-grid", Container)
        insights = self.query_one("#insp-insights", Static)
        preview = self.query_one("#insp-preview_content", Static)
        raw = self.query_one("#insp-raw_content", Static)
        
        # Clear all
        details_grid.remove_children()
        insights.update("")
        preview.update("")
        raw.update("")
        
        if node is None:
            path_view.update("")
            return

        # Path Header
        chain = []
        curr = node
        while curr:
            label = curr.key if curr.key else "root"
            chain.append(label)
            if curr.parent:
                curr = self.model.get_node(curr.parent)
            else:
                curr = None
        human_path = " › ".join(reversed(chain))
        path_view.update(human_path)

        # Details Grid
        type_str = node.type.value.capitalize()
        size_str = "-"
        
        children_count = 0
        if node.type in (DataType.ARRAY, DataType.OBJECT):
             children_count = self.model.get_children_count(node.id)
             size_str = f"{children_count} items"
        elif node.type == DataType.STRING:
             size_str = f"{len(str(node.value))} chars"

        def add_detail(label: str, value: str):
            details_grid.mount(Label(label, classes="insp-label"))
            details_grid.mount(Label(value, classes="insp-value"))

        add_detail("Type", type_str)
        add_detail("Size", size_str)
        
        # Smart Insights 
        insight_text = Text()
        has_insights = False
        
        if node.type == DataType.STRING:
            val_str = str(node.value)
            
            # URL
            if self.URL_PATTERN.match(val_str):
                add_detail("Format", "URL")
                insight_text.append("\nLink detected:\n", style="bold")
                insight_text.append(val_str, style="blue underline link " + val_str)
                has_insights = True
            
            # Color
            elif self.HEX_COLOR_PATTERN.match(val_str):
                add_detail("Format", "Color")
                insight_text.append("\nColor Preview: ", style="bold")
                insight_text.append("██████", style=val_str)
                insight_text.append(f" {val_str}")
                has_insights = True
                
            # Date
            else:
                try:
                    dt = datetime.fromisoformat(val_str.replace('Z', '+00:00'))
                    add_detail("Format", "ISO8601")
                    add_detail("Time", dt.strftime('%Y-%m-%d %H:%M'))
                except ValueError:
                    pass
        
        if has_insights:
             insights.update(Panel(insight_text, title="Smart Insights", border_style="dim"))
        
        # Preview / Raw
        if node.is_container:
            # Preview: Show first 20 items simply
            preview_str = f""
            
            children = self.model.get_children(node.id)
            
            limit = 30
            for i, child in enumerate(children[:limit]): 
                icon = "[+] " if child.is_container else "- "
                val = "..." if child.is_container else str(child.value)
                if len(val) > 50: val = val[:47] + "..."
                preview_str += f"{icon} {child.key}: {val}\n"
            
            if children_count > limit:
                preview_str += f"\n... and {children_count - limit} more."
                
            preview.update(Panel(preview_str, title="Content Preview"))
            
            # Raw View: Attempt to load if size is reasonable
            # 500 items is a safe upper bound for UI responsiveness
            if children_count < 500:
                try:
                    # Depth 4 is enough for inspection without exploding
                    data = self.model.reconstruct_json(node.id, max_depth=4)
                    
                    if self.format == "yaml" and yaml:
                        # Serialize as YAML
                        source_str = yaml.dump(data, sort_keys=False, default_flow_style=False)
                        lexer = "yaml"
                    else:
                        # Serialize as JSON
                        source_str = json.dumps(data, indent=2)
                        lexer = "json"

                    raw.update(Panel(Syntax(source_str, lexer, theme=self.SYNTAX_THEME, padding=1), title="Source"))
                except Exception as e:
                    raw.update(Panel(Text(f"Error loading raw view: {e}", style="red"), title="Source"))
            else:
                raw.update(Panel(Text("Raw view hidden for performance ( > 500 items)", style="italic dim"), title="Source"))
            
        else:
            # Primitive
            preview.update(Panel(str(node.value), title="Value"))
            
            # Raw: Try to parse as JSON if it looks like it, else just syntax highlight the value
            if self.format == "yaml" and yaml:
                # For primitives, YAML is just the value often, or specialized
                 source_str = yaml.dump(node.value, sort_keys=False)
                 lexer = "yaml"
            else:
                 source_str = json.dumps(node.value, indent=2)
                 lexer = "json"

            raw.update(Panel(Syntax(source_str, lexer, theme=self.SYNTAX_THEME, padding=1), title="Source"))



import sqlite3
import sys
import os
import hashlib
from pathlib import Path
from typing import Optional

class DatabaseManager:
    """
    Manages SQLite database connections and file locations.
    Uses a cache directory to store database files hashed by the input file path.
    """
    
    def __init__(self, debug: bool = False):
        self.debug = debug
        self._cache_dir = self._get_cache_dir()
        self._cache_dir.mkdir(parents=True, exist_ok=True)

    def _get_cache_dir(self) -> Path:
        """
        Returns the platform-specific cache directory.
        """
        if sys.platform == "win32":
            local_app_data = os.environ.get("LOCALAPPDATA")
            if local_app_data:
                return Path(local_app_data) / "twig"
            return Path.home() / "AppData" / "Local" / "twig"
        elif sys.platform == "darwin":
            return Path.home() / "Library" / "Caches" / "twig"
        else:
            # Linux / XDG
            xdg_cache = os.environ.get("XDG_CACHE_HOME")
            if xdg_cache:
                return Path(xdg_cache) / "twig"
            return Path.home() / ".cache" / "twig"

    def get_db_path(self, input_file_path: str) -> Path:
        """
        Generates a deterministic database file path for a given input file.
        Hash includes the absolute path and modification time/size could allow invalidation,
        but for now we just hash the path.
        
        To handle file updates, the loader should check mtime or we just overwrite if requested.
        """
        abs_path = str(Path(input_file_path).resolve())
        # Use simple md5 for filename
        file_hash = hashlib.md5(abs_path.encode("utf-8")).hexdigest()
        filename = f"{Path(input_file_path).name}_{file_hash}.db"
        return self._cache_dir / filename

    def get_connection(self, db_path: Path) -> sqlite3.Connection:
        """
        Returns a configured sqlite3 connection.
        """
        conn = sqlite3.connect(str(db_path), check_same_thread=False)
        conn.row_factory = sqlite3.Row
        
        # Performance tuning for bulk loads and reads
        conn.execute("PRAGMA journal_mode = WAL;")  # Write-Ahead Logging
        conn.execute("PRAGMA synchronous = NORMAL;") 
        conn.execute("PRAGMA temp_store = MEMORY;")
        conn.execute("PRAGMA cache_size = -64000;") # ~64MB cache
        
        return conn

    def init_db(self, db_path: Path) -> None:
        """
        Initializes the database schema.
        """
        schema_path = Path(__file__).parent / "schema.sql"
        if not schema_path.exists():
            raise FileNotFoundError("schema.sql not found")
            
        with open(schema_path, "r", encoding="utf-8") as f:
            schema_sql = f.read()
            
        conn = self.get_connection(db_path)
        try:
            conn.executescript(schema_sql)
            conn.commit()
        finally:
            conn.close()

    def clear_cache(self) -> None:
        """
        Clears all cached database files.
        """
        for file in self._cache_dir.glob("*.db"):
            try:
                file.unlink()
            except OSError:
                pass
                
        for file in self._cache_dir.glob("*.db-wal"):
            try:
                file.unlink()
            except OSError:
                pass
                
        for file in self._cache_dir.glob("*.db-shm"):
            try:
                file.unlink()
            except OSError:
                pass



from textual.app import ComposeResult
from textual.containers import HorizontalScroll, Vertical
from textual.widgets import OptionList
from textual.message import Message
from textual.binding import Binding
import uuid
from typing import List, Optional
import asyncio
import re
from rich.text import Text

from twg.core.model import SQLiteModel, Node, DataType

class TwigOptionList(OptionList):
    """
    A customized OptionList that handles focus movement events.
    """
    BINDINGS = [
        Binding("left", "move_focus(-1)", "Back", show=False),
        Binding("right", "move_focus(1)", "Forward", show=False),
    ]

    class MoveFocus(Message):
        def __init__(self, direction: int):
            self.direction = direction
            super().__init__()

    def action_move_focus(self, direction: int) -> None:
        self.post_message(self.MoveFocus(direction))

class Column(Vertical):
    """A single column in the Miller Column view."""
    
    class Highlighted(Message):
        def __init__(self, column_index: int, node_id: uuid.UUID):
            self.column_index = column_index
            self.node_id = node_id
            super().__init__()

    def __init__(self, model: SQLiteModel, parent_node_id: uuid.UUID, index: int, initial_select_index: int = 0):
        self.model = model
        self.parent_node_id = parent_node_id
        self.index = index
        self.initial_select_index = initial_select_index
        super().__init__(classes="column")

    def compose(self) -> ComposeResult:
        try:
            parent = self.model.get_node(self.parent_node_id)
            if not parent:
                return

            options = []
            self.node_map = {} # Map option index to node ID
            
            # Lazy load children
            children = self.model.get_children(parent.id)
            
            for i, child in enumerate(children):
                if child:
                    # Icon logic
                    icon = " "
                    if child.type == DataType.OBJECT:
                        icon = "▶" 
                    elif child.type == DataType.ARRAY:
                        icon = "▶" 
                    
                    # Value preview (truncated)
                    val_str = ""
                    if not child.is_container:
                        val_str = f": {str(child.value)}"
                        if len(val_str) > 20:
                            val_str = val_str[:17] + "..."
                    
                    label = f"{icon} {child.key}{val_str}"
                    options.append(label)
                    self.node_map[i] = child.id
            
            opt_list = TwigOptionList(id=f"col-{self.index}")
            opt_list.add_options(options)
            yield opt_list
            
        except Exception as e:
             from textual.widgets import Label
             yield Label(f"Error loading column: {e}")

    def on_mount(self) -> None:
        """Select the initial item by default when the column is mounted."""
        # Check if list exists (might have failed to load)
        if not self.query(TwigOptionList):
            return

        opt_list = self.query_one(TwigOptionList)
        if opt_list.option_count > 0:
            # Respect the initial requested selection
            target = self.initial_select_index if self.initial_select_index < opt_list.option_count else 0
            opt_list.highlighted = target
            opt_list.scroll_to_highlight()

    def on_option_list_option_highlighted(self, event: OptionList.OptionHighlighted) -> None:
        has_focus = self.query_one(TwigOptionList).has_focus
        
        if has_focus:
            node_id = self.node_map.get(event.option_index)
            if node_id:
                self.post_message(self.Highlighted(self.index, node_id))
        else:
            pass # No focus, ignore

    def highlight_text_match(self, index: int, query: str) -> None:
        """
        Highlights the query text within the option at the given index.
        """
        opt_list = self.query_one(TwigOptionList)
        if index < 0 or index >= opt_list.option_count:
            return

        node_id = self.node_map.get(index)
        if not node_id: return
        
        node = self.model.get_node(node_id)
        if not node: return
        
        # Reconstruct label logic (duplicate of compose, but needed for Text construction)
        icon = " "
        if node.type == DataType.OBJECT:
            icon = "▶" 
        elif node.type == DataType.ARRAY:
            icon = "▶" 
        
        val_str = ""
        if not node.is_container:
            val_str = f": {str(node.value)}"
            if len(val_str) > 20:
                val_str = val_str[:17] + "..."
        
        full_label = f"{icon} {node.key}{val_str}"
        
        text = Text(full_label)
        if query:
            pattern = re.compile(re.escape(query), re.IGNORECASE)
            for match in pattern.finditer(full_label):
                text.stylize("bold reverse yellow", match.start(), match.end())
        
        try:
            opt_list.replace_option_at_index(index, text)
        except AttributeError:
             pass

    # find_next removed as it was unused and global search is preferred.



class ColumnNavigator(HorizontalScroll):
    """
    Manages the Miller Columns navigation.
    
    Handles column creation, removal, and focus management as the user navigates
    through the JSON structure.
    """

    class NodeSelected(Message):
        def __init__(self, node_id: uuid.UUID):
            self.node_id = node_id
            super().__init__()

    def __init__(self, model: SQLiteModel):
        self.model = model
        self._nav_id = 0 # To handle cancellation of jump/expand tasks
        self._pending_target_node_id: Optional[uuid.UUID] = None
        self._is_expanding = False
        super().__init__()

    def compose(self) -> ComposeResult:
        # Start with the root column
        if self.model.root_id:
            yield Column(self.model, self.model.root_id, 0)

    def __arrow_up(self) -> None:
        pass # Optional hook override

    async def _expand_node(self, column_index: int, node_id: uuid.UUID, initial_select_index: int = 0, animate: bool = True) -> None:
        """Helper to expand a node into a new column with a specific selection."""
        next_col_index = column_index + 1
        existing_next_col = None
        
        for child in self.children:
            if isinstance(child, Column) and child.index == next_col_index:
                existing_next_col = child
                break
        
        # If the next column exists and is showing the children of OUR selected node, we are good.
        if existing_next_col and existing_next_col.parent_node_id == node_id:
             self.post_message(self.NodeSelected(node_id))
             return

        to_remove = [child for child in self.children if isinstance(child, Column) and child.index > column_index]
        if to_remove:
             await asyncio.gather(*[child.remove() for child in to_remove])
        
        node = self.model.get_node(node_id)
        if node and node.is_container:
            for child in self.children:
                if isinstance(child, Column) and child.index == next_col_index:
                    await child.remove()

            new_col = Column(self.model, node.id, next_col_index, initial_select_index=initial_select_index)
            await self.mount(new_col)
            
            if animate:
                self.call_after_refresh(self.scroll_to_widget, new_col, animate=True)
            else:
                self.scroll_to_widget(new_col, animate=False)

        self.post_message(self.NodeSelected(node_id))

    async def expand_to_node(self, node_id: uuid.UUID) -> None:
        """
        Expands the column view to show and select the given node.
        Rebuilds the navigation path from root to the target node.
        """
        # Increment generation to cancel any running expansion
        self._nav_id += 1
        current_id = self._nav_id
        self._pending_target_node_id = node_id
        self._is_expanding = True
        
        try:
            node = self.model.get_node(node_id)
            if not node: return
            
            lineage = []
            curr = node
            while curr:
                lineage.append(curr)
                curr = self.model.get_node(curr.parent) if curr.parent else None
            lineage.reverse()
        
            for i in range(len(lineage) - 1):
                 if self._nav_id != current_id:
                     return

                 current_node = lineage[i]
                 next_node = lineage[i+1] # The one selected in Col i
                 
                 try:
                     children = self.model.get_children(current_node.id)
                     
                     select_idx = -1
                     for idx, child in enumerate(children):
                         if child.id == next_node.id:
                             select_idx = idx
                             break
                     
                     if select_idx != -1:
                         grandchild_idx = 0
                         if i + 2 < len(lineage):
                             grandchild = lineage[i+2]
                             grand_children = self.model.get_children(next_node.id)
                             for gc_idx, gc in enumerate(grand_children):
                                 if gc.id == grandchild.id:
                                     grandchild_idx = gc_idx
                                     break
                         
                         col_widget = None
                         for child in self.children:
                             if isinstance(child, Column) and child.index == i:
                                 col_widget = child
                                 break
                         
                         if col_widget:
                             if not col_widget.query(TwigOptionList):
                                 continue
                                 
                             opt_list = col_widget.query_one(TwigOptionList)
                             opt_list.highlighted = select_idx
                             opt_list.scroll_to_highlight(top=True) # Pin to top for predictable jumping
                             self.scroll_to_widget(col_widget, animate=False)
                             
                             await self._expand_node(i, next_node.id, initial_select_index=grandchild_idx, animate=False)
                             
                 except Exception as e:
                     self.notify(f"Jump Error: {e}", severity="error")
                     break
        
        finally:
             self._is_expanding = False

        if self._nav_id != current_id:
            return
            
        self._pending_target_node_id = None

        final_col_idx = len(lineage) - 1
        target_col_idx = max(0, len(lineage) - 2)
        
        def focus_target():
             if self._nav_id != current_id: return
             
             for child in self.children:
                 if isinstance(child, Column) and child.index == target_col_idx:
                     if child.query(TwigOptionList):
                         child.query_one(TwigOptionList).focus()
                         self.post_message(self.NodeSelected(node_id))
                     break
        self.call_after_refresh(focus_target)

    async def find_next(self, query: str, direction: int = 1) -> Optional[Node]:
        """
        Global search using the model's DFS.
        Returns the Node found, or None.
        """
        # 1. Determine start Node ID from current focus
        start_node_id = self._pending_target_node_id
        focused_col = None
        
        if not start_node_id:
            # Try finding focused column first
            for child in self.children:
                 if isinstance(child, Column) and child.query(TwigOptionList):
                     if child.query_one(TwigOptionList).has_focus:
                         focused_col = child
                         break
            
            # Fallback: Find deepest column with a highlight (if focus lost due to Loading modal)
            if not focused_col:
                 max_idx = -1
                 for child in self.children:
                     if isinstance(child, Column) and child.query(TwigOptionList):
                         opts = child.query_one(TwigOptionList)
                         if opts.highlighted is not None and child.index > max_idx:
                             max_idx = child.index
                             focused_col = child

            if focused_col and focused_col.query(TwigOptionList):
                opts = focused_col.query_one(TwigOptionList)
                if opts.highlighted is not None:
                    start_node_id = focused_col.node_map.get(opts.highlighted)
        
        # 2. Search in Model
        target_node = None
        
        # Feature: Smart Search (if query looks like a path, try resolving it first)
        if query.startswith("."):
            try:
                resolved = self.model.resolve_path(query)
                if resolved:
                    target_node = resolved
            except:
                pass

        if not target_node:
             # Note: Direction ignored for now, always forward global search since we use DFS generator
             target_node = self.model.find_next_node(query, start_node_id=start_node_id, direction=direction)
        
        if target_node:
            # 3. Expand to it
            await self.expand_to_node(target_node.id)
            # Post selection
            self.post_message(self.NodeSelected(target_node.id))
            
            # 4. Highlight text match in the target column
            for child in self.children:
                if isinstance(child, Column) and child.parent_node_id == target_node.parent:
                     if child.query(TwigOptionList):
                         opts = child.query_one(TwigOptionList)
                         if opts.highlighted is not None:
                             child.highlight_text_match(opts.highlighted, query)
            
            return target_node
        
        return None

    async def on_mount(self) -> None:
        # Focus the first option list
        if self.children:
            first_col = self.children[0]
            if isinstance(first_col, Column) and first_col.query(TwigOptionList):
                opt_list = first_col.query_one(TwigOptionList)
                opt_list.focus()
                # Manually trigger initial selection
                if opt_list.highlighted is not None:
                     node_id = first_col.node_map.get(opt_list.highlighted)
                     if node_id:
                         await self._expand_node(first_col.index, node_id, initial_select_index=0)

    async def on_column_highlighted(self, message: Column.Highlighted) -> None:
        """Handles selection within a column."""
        if self._is_expanding:
            return
        await self._expand_node(message.column_index, message.node_id, initial_select_index=0)

    def on_twig_option_list_move_focus(self, message: TwigOptionList.MoveFocus) -> None:
        self._move_focus(message.direction)

    def _move_focus(self, direction: int) -> None:
        # Find currently focused column index
        focused = self.screen.focused
        if not focused:
            return

        # Find which column contains the focused widget
        current_col = None
        for child in self.children:
            if isinstance(child, Column) and child.query_one(TwigOptionList).has_focus:
                current_col = child
                break
        
        if current_col:
            target_index = current_col.index + direction
            # Find target column
            for child in self.children:
                if isinstance(child, Column) and child.index == target_index:
                    target_col = child
                    if not target_col.query(TwigOptionList):
                        self.notify("Cannot focus empty column", severity="warning")
                        continue
                        
                    target_option_list = target_col.query_one(TwigOptionList)
                    target_option_list.focus()
                    self.scroll_to_widget(target_col, animate=True)
                    
                    if target_option_list.highlighted is not None:
                        node_id = target_col.node_map.get(target_option_list.highlighted)
                        if node_id:
                            self.run_worker(self.on_column_highlighted(Column.Highlighted(target_col.index, node_id)))
                    return



from textual.widgets import Static
from textual.reactive import reactive
from twg.core.model import Node, SQLiteModel

class Breadcrumbs(Static):
    """
    Displays the path to the currently selected node.
    
    Uses jq-style syntax (e.g., .users[0].name) for the path.
    """
    # CSS moved to styles.tcss

    selected_node: reactive[Node | None] = reactive(None)

    def __init__(self, model: SQLiteModel, **kwargs):
        self.model = model
        super().__init__(**kwargs)

    def watch_selected_node(self, node: Node | None) -> None:
        if not node:
            self.update("Path: .")
            return
        
        full_path = self.model.get_path(node.id)
        
        # Smart truncation for long paths
        if len(full_path) > 100:
            parts = full_path.split('.')
            if len(parts) > 6:
                # Keep first 2 and last 3 for better context
                full_path = f"{parts[0]}.{parts[1]} ... {parts[-3]}.{parts[-2]}.{parts[-1]}"
            else:
                # Fallback simple truncation - preserve end
                full_path = "..." + full_path[-97:]

        self.update(f"[b]Path:[/b] {full_path} [dim](jq)[/]")



import os
import json
import sys
from pathlib import Path
from typing import Any, Dict

class Config:
    """
    Handles loading and saving of application configuration.
    
    Configuration is stored in a JSON file in the standard user configuration directory:
    - Windows: %APPDATA%/twig/config.json
    - macOS/Linux: $XDG_CONFIG_HOME/twig/config.json (defaults to ~/.config/twig/config.json)
    """
    
    DEFAULT_CONFIG = {
        "theme": "catppuccin-mocha"
    }

    def __init__(self):
        self._config_dir = self._get_config_dir()
        self._config_file = self._config_dir / "config.json"
        self._data: Dict[str, Any] = self.DEFAULT_CONFIG.copy()
        self.load()

    def _get_config_dir(self) -> Path:
        """
        Returns the platform-specific configuration directory.
        """
        if sys.platform == "win32":
            app_data = os.environ.get("APPDATA")
            if app_data:
                return Path(app_data) / "twig"
            return Path.home() / "AppData" / "Roaming" / "twig"
        else:
            # macOS and Linux
            xdg_config = os.environ.get("XDG_CONFIG_HOME")
            if xdg_config:
                return Path(xdg_config) / "twig"
            return Path.home() / ".config" / "twig"

    def load(self) -> None:
        """
        Loads the configuration from disk. 
        If the file doesn't exist, keeps the default config.
        """
        if not self._config_file.exists():
            return

        try:
            with open(self._config_file, "r", encoding="utf-8") as f:
                user_config = json.load(f)
                # Update defaults with user config (merging)
                self._data.update(user_config)
        except (OSError, json.JSONDecodeError) as e:
            print(f"Warning: Failed to load config file: {e}", file=sys.stderr)

    def save(self) -> None:
        """
        Saves the current configuration to disk.
        Creates the directory if it doesn't exist.
        """
        try:
            self._config_dir.mkdir(parents=True, exist_ok=True)
            with open(self._config_file, "w", encoding="utf-8") as f:
                json.dump(self._data, f, indent=4)
        except OSError as e:
             print(f"Warning: Failed to save config file: {e}", file=sys.stderr)

    def get(self, key: str, default: Any = None) -> Any:
        """Get a configuration value."""
        return self._data.get(key, default)

    def set(self, key: str, value: Any) -> None:
        """Set a configuration value and save immediately."""
        self._data[key] = value
        self.save()



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Generator
import uuid
import re
import sqlite3
from twg.core.db import DatabaseManager

class DataType(Enum):
    OBJECT = "object"
    ARRAY = "array"
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    BOOLEAN = "boolean"
    NULL = "null"
    
    @staticmethod
    def from_value(data: Any) -> DataType:
        if data is None:
            return DataType.NULL
        elif isinstance(data, bool):
            return DataType.BOOLEAN
        elif isinstance(data, int):
            return DataType.INTEGER
        elif isinstance(data, float):
            return DataType.FLOAT
        elif isinstance(data, str):
            return DataType.STRING
        elif isinstance(data, dict):
            return DataType.OBJECT
        elif isinstance(data, list):
            return DataType.ARRAY
        return DataType.STRING

@dataclass
class Node:
    """
    Lightweight Node representation from SQLite.
    Does not store children or raw values in memory.
    """
    id: uuid.UUID
    key: str
    value: Any
    type: DataType
    parent: Optional[uuid.UUID]
    path: str
    is_expanded: bool = False
    
    @property
    def is_container(self) -> bool:
        return self.type in (DataType.OBJECT, DataType.ARRAY)

class SQLiteModel:
    """
    SQLite-backed data model.
    """
    def __init__(self, db_path: str):
        self.db_manager = DatabaseManager()
        self.db_path = db_path
        # We keep a connection for read operations
        self.conn = self.db_manager.get_connection(self.db_path)
        
        # Cache root id to avoid query
        self.root_id = self._fetch_root_id()

    def _fetch_root_id(self) -> Optional[uuid.UUID]:
        cursor = self.conn.execute("SELECT id FROM nodes WHERE parent_id IS NULL LIMIT 1")
        row = cursor.fetchone()
        return uuid.UUID(row["id"]) if row else None

    def close(self):
        self.conn.close()

    def row_to_node(self, row: sqlite3.Row) -> Node:
        # Convert type string back to Enum
        try:
             # simple lookup by value
             dtype = DataType(row["type"])
        except ValueError:
             dtype = DataType.STRING
             
        # Deserialize value from DB (stored as TEXT)
        # We attempt to cast back to native types where appropriate
        val = row["value"]
        if dtype == DataType.NULL:
            val = None
        elif dtype == DataType.BOOLEAN:
            val = (val == "True" or val == "true")
        elif dtype == DataType.INTEGER:
            try: val = int(val)
            except: pass
        elif dtype == DataType.FLOAT:
             try: val = float(val)
             except: pass
             
        return Node(
            id=uuid.UUID(row["id"]),
            key=row["key"],
            value=val,
            type=dtype,
            parent=uuid.UUID(row["parent_id"]) if row["parent_id"] else None,
            path=row["path"],
            is_expanded=bool(row["is_expanded"])
        )

    def get_node(self, node_id: uuid.UUID) -> Optional[Node]:
        cursor = self.conn.execute(
            "SELECT * FROM nodes WHERE id = ?", 
            (str(node_id),)
        )
        row = cursor.fetchone()
        if row:
            return self.row_to_node(row)
        return None

    def get_children(self, node_id: uuid.UUID) -> List[Node]:
        """Returns ordered children of the node."""
        cursor = self.conn.execute(
            "SELECT * FROM nodes WHERE parent_id = ? ORDER BY rank", 
            (str(node_id),)
        )
        return [self.row_to_node(row) for row in cursor.fetchall()]

    def get_children_count(self, node_id: uuid.UUID) -> int:
        """Returns the number of children."""
        cursor = self.conn.execute(
            "SELECT COUNT(*) FROM nodes WHERE parent_id = ?", 
            (str(node_id),)
        )
        return cursor.fetchone()[0]

    def get_path(self, node_id: uuid.UUID) -> str:
        """Returns the jq-style path from the DB column."""
        node = self.get_node(node_id)
        if node:
            return node.path
        return ""

    def find_next_node(self, query: str, start_node_id: Optional[uuid.UUID] = None, direction: int = 1) -> Optional[Node]:
        """
        Global search using substring match (LIKE %query%).
        
        Searches 'key' and 'value' columns.
        Results are ordered by path to provide a consistent navigation order.
        """
        query = query.strip()
        if not query: return None
        
        # Substring pattern
        like_pattern = f"%{query}%"
        start_path = None

        if start_node_id:
            n = self.get_node(start_node_id)
            if n:
                start_path = n.path
        
        # We search KEY and VALUE columns.
        # Note: 'value' column stores primitives as strings.
        
        if start_path:
            if direction > 0:
                 # Find next match after current path
                 sql_next = """
                    SELECT * FROM nodes 
                    WHERE (key LIKE ? OR value LIKE ?) AND path > ?
                    ORDER BY path ASC LIMIT 1
                 """
                 cursor = self.conn.execute(sql_next, (like_pattern, like_pattern, start_path))
                 row = cursor.fetchone()
                 if row: return self.row_to_node(row)
                 
                 # Wrap around: Find first match from start
                 sql_first = """
                    SELECT * FROM nodes 
                    WHERE (key LIKE ? OR value LIKE ?)
                    ORDER BY path ASC LIMIT 1
                 """
                 cursor = self.conn.execute(sql_first, (like_pattern, like_pattern))
                 row = cursor.fetchone()
                 if row: return self.row_to_node(row)
                 
            else:
                 # Find prev match before current path
                 sql_prev = """
                    SELECT * FROM nodes 
                    WHERE (key LIKE ? OR value LIKE ?) AND path < ?
                    ORDER BY path DESC LIMIT 1
                 """
                 cursor = self.conn.execute(sql_prev, (like_pattern, like_pattern, start_path))
                 row = cursor.fetchone()
                 if row: return self.row_to_node(row)
                 
                 # Wrap around: Find last match
                 sql_last = """
                    SELECT * FROM nodes 
                    WHERE (key LIKE ? OR value LIKE ?)
                    ORDER BY path DESC LIMIT 1
                 """
                 cursor = self.conn.execute(sql_last, (like_pattern, like_pattern))
                 row = cursor.fetchone()
                 if row: return self.row_to_node(row)
                 
        else:
             # No context, just return first
             sql_first = """
                SELECT * FROM nodes 
                WHERE (key LIKE ? OR value LIKE ?)
                ORDER BY path ASC LIMIT 1
             """
             cursor = self.conn.execute(sql_first, (like_pattern, like_pattern))
             row = cursor.fetchone()
             if row: return self.row_to_node(row)

        return None

    def resolve_path(self, path: str) -> Optional[Node]:
        """
        Direct lookup by path column.
        """
        path = path.strip()
        if not path: return None
        if not path.startswith("."): 
             # Normalize paths to always start with dot (except empty, handled above)
             path = "." + path 

        # Exact match
        cursor = self.conn.execute("SELECT * FROM nodes WHERE path = ?", (path,))
        row = cursor.fetchone()
        if row:
            return self.row_to_node(row)
            
        # Smart Fallback for single-document YAMLs (wrapped in Stream array)
        # If user types .kind, but it's actually .[0].kind
        if path.startswith("."):
            fallback_path = ".[0]" + path[1:] # .kind -> .[0].kind
            cursor = self.conn.execute("SELECT * FROM nodes WHERE path = ?", (fallback_path,))
            row = cursor.fetchone()
            if row:
                return self.row_to_node(row)

        return None

    def get_search_stats(self, query: str, current_node_id: Optional[uuid.UUID]) -> tuple[int, int]:
        """
        Returns (current_match_index, total_matches) for the given query.
        Index is 1-based. Returns (0, 0) if no matches or query empty.
        """
        query = query.strip()
        if not query: return (0, 0)
        
        like_pattern = f"%{query}%"
        
        sql_count = "SELECT COUNT(*) FROM nodes WHERE key LIKE ? OR value LIKE ?"
        cursor = self.conn.execute(sql_count, (like_pattern, like_pattern))
        total = cursor.fetchone()[0]
        
        if total == 0:
            return (0, 0)
            
        current_index = 0
        if current_node_id:
             node = self.get_node(current_node_id)
             if node:
                 # Count how many matches have a path <= current node's path
                 sql_rank = """
                    SELECT COUNT(*) 
                    FROM nodes 
                    WHERE (key LIKE ? OR value LIKE ?) AND path <= ?
                 """
                 cursor = self.conn.execute(sql_rank, (like_pattern, like_pattern, node.path))
                 current_index = cursor.fetchone()[0]
                 
        return (current_index, total)

    def reconstruct_json(self, node_id: uuid.UUID, max_depth: int = 3, current_depth: int = 0) -> Any:
        """
        Reconstructs the JSON value for a given node.
        Limits depth to prevent excessive loading.
        """
        node = self.get_node(node_id)
        if not node: return None
        
        if not node.is_container:
            return node.value
            
        if current_depth >= max_depth:
            return "..."
            
        children = self.get_children(node_id)
        
        if node.type == DataType.OBJECT:
            res = {}
            for child in children:
                res[child.key] = self.reconstruct_json(child.id, max_depth, current_depth + 1)
            return res
            
        elif node.type == DataType.ARRAY:
            res = []
            for child in children:
                res.append(self.reconstruct_json(child.id, max_depth, current_depth + 1))
            return res
            
        return None
